/*
 * Copyright (c) 2003-2021 JADAPTIVE Limited. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code and
 * its use is subject to the terms of the JADAPTIVE Software License. You may not use
 * this file except in compliance with the license terms.
 *
 * https://www.jadaptive.com/en/content/api-software-license
 *
 * Email: support@jadaptive.com
 * WWW: https://www.jadaptive.com
 */
import static com.maverick.sftp.SftpSubsystemChannel.OPEN_CREATE;
import static com.maverick.sftp.SftpSubsystemChannel.OPEN_EXCLUSIVE;
import static com.maverick.sftp.SftpSubsystemChannel.OPEN_WRITE;
import static java.lang.System.exit;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Deque;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;

import com.maverick.sftp.SftpFileOutputStream;
import com.maverick.sftp.SftpStatusException;
import com.maverick.ssh.ChannelOpenException;
import com.maverick.ssh.PasswordAuthentication;
import com.maverick.ssh.SshAuthentication;
import com.maverick.ssh.SshClient;
import com.maverick.ssh.SshConnector;
import com.maverick.ssh.SshException;
import com.maverick.ssh2.Ssh2Client;
import com.maverick.ssh2.Ssh2Context;
import com.maverick.ssh2.Ssh2Session;
import com.sshtools.net.SocketTransport;
import com.sshtools.sftp.SftpClient;

/**
 * Created by ubeinges on 01.07.14.
 */
public class MaverickTests {

    private static DatatypeFactory DATATYPE_FACTORY;

    static {
        setLicense();

        try {
            DATATYPE_FACTORY = DatatypeFactory.newInstance();
        } catch (DatatypeConfigurationException e) {
            throw new IllegalStateException(e.getMessage(), e);
        }
    }

    /**
     * SFTP timeout in milliseconds.
     * Can be configured via the system property ima.archive.sftp.timeout.
     * Default value is 60000 (one minute).
     */
    protected static final int TIMEOUT =
            Integer.getInteger("ima.archive.sftp.timeout", 5000);

    /**
     * Preferred public key type for J2SSH.
     * Can be configured via the system property ima.sftp.j2ssh.preferredPublicKeyType.
     * For allowed values see {@link Ssh2Context}.
     * The default is set to RSA since normal known_hosts file usually uses this.
     */
    private static final String PREFERRED_PUBLIC_KEY_TYPE =
            System.getProperty("_ima.sftp.j2ssh.preferredPublicKeyType", Ssh2Context.PUBLIC_KEY_SSHRSA);

    /**
     * Set alternate TCP buffersize.
     * Can be configured via the system property ima.sftp.j2ssh.tcpBufferSize.
     * If not set or set to 0, the system default will be used.
     */
    private static final int TCP_BUFFERSIZE =
            Integer.getInteger("_ima.sftp.j2ssh.tcpBufferSize", 0);

    /**
     * Set alternate TCP_NODELAY option.
     * If set to true the Nagle algorithm will be disabled.
     * Can be configured via the system property ima.sftp.j2ssh.tcpNoDelay.
     * If not set, the system default of false will be used.
     */
    private static final boolean TCP_NODELAY =
            Boolean.getBoolean("_ima.sftp.j2ssh.tcpNoDelay");

    /**
     * Set alternate window size for sftp.
     * This will tune the initial window size used for ssh2 sessions for SFTP.
     * Can be configured via the system property ima.sftp.j2ssh.windowSize.
     * If not set, the system default of 1m will be used.
     */
    private static final int WINDOWSIZE =
            Integer.getInteger("_ima.sftp.j2ssh.windowSize", 1024*1024);

    /**
     * Set alternate packet size for sftp.
     * This will tune the maximum packet size used for ssh2 sessions for SFTP.
     * Can be configured via the system property ima.sftp.j2ssh.packetSize.
     * If not set, the system default of 16k will be used.
     */
    private static final int PACKETSIZE =
            Integer.getInteger("_ima.sftp.j2ssh.packetSize", 1024*16);

    /** set the desired cipher */
    private static final String CIPHER =
            System.getProperty("_ima.sftp.j2ssh.cipher", "aes128-ctr");

    /** set the desired mac */
    private static final String MAC =
            System.getProperty("_ima.sftp.j2ssh.mac", "hmac-md5");

    /** size of the test content */
    private static final int CONTENT_SIZE = 1 * 1024;
    /** buffer size for read/write */
    private static final int BUFFER_SIZE = 8 * 1024;
    /** sleep time for dummy worker */
    private static long DUMMY_SLEEP_MILLIS = 100;

    private static PoolingScope POOLING_SCOPE = PoolingScope.PROCESS;
    private static WorkloadType WORKLOAD_TYPE = WorkloadType.NORMAL;

    private static String HOST;
    private static int PORT;
    private static String USER;
    private static String PASS;
    private static String ROOT_DIRECTORY = "Maverick-Tests";
    private static int DEBUG_LEVEL = 0;

    private enum PoolingScope {
        NONE,
        PROCESS,
        PROCESS_FIFO,
        THREAD
    }

    private enum WorkloadType {
        NORMAL,
        NONE,
        DUMMY
    }

    private static void usage(String programName) {

        System.out.println("Usage: " + programName + " [options] <host> <port> <user> <password> <#threads> <#files-per-thread>");
        System.out.println("Options:");
        System.out.println("  -h, --help:                   Print usage information.");
        System.out.println("  --debug <level>:              Set debug level. You can set a level from 0 (least/off) to 9 (most).");
        System.out.println("                                Default is not set is " + DEBUG_LEVEL + ".");
        System.out.println("  --workload-type <type>:       Set the desired type of workload.");
        System.out.println("                                Available types are {NORMAL, NONE, DUMMY}.");
        System.out.println("                                Default is not set is " + WORKLOAD_TYPE + ".");
        System.out.println("                                NORMAL writes some, files, reads them and deletes them afterwards.");
        System.out.println("                                NONE does nothing, only acquires and releases the connection.");
        System.out.println("                                DUMMY acquires the connection, sleeps for some time releases the connection.");
        System.out.println("  --dummy-sleep-time <millis>:  Set the sleep milliseconds for the DUMMY workload.");
        System.out.println("                                Default is not set is " + DUMMY_SLEEP_MILLIS + ".");
        System.out.println("  --pooling-scope <scope>:      Set desired scope for connection pooling.");
        System.out.println("                                Available scopes are {NONE, PROCESS, THREAD}.");
        System.out.println("                                Default is not set is " + POOLING_SCOPE + ".");
        System.out.println("                                PROCESS sets the pooling scope to the java process.");
        System.out.println("                                THREAD sets the pooling scope to the java thread.");
        System.out.println("                                NONE will disable connection pooling.");
    }

    public static void main(String[] args) {

        CommandParser parser = new CommandParser(args, "workload-type:", "dummy-sleep-millis:", "pooling-scope:", "help", "h", "debug:");

        List<String> parsedArgs = parser.getArgs();

        System.out.println("parsedArgs = " + parsedArgs);

        if (parser.hasOption("help") || parser.hasOption("h")) {
            usage("MaverickTests");
            exit(1);
        }

        if (parser.hasOption("debug")) {
            DEBUG_LEVEL = new Integer(parser.getOptionValue("debug")).intValue();
        }

        if (parser.hasOption("dummy-sleep-millis")) {
            DUMMY_SLEEP_MILLIS = new Long(parser.getOptionValue("dummy-sleep-millis")).longValue();
        }

        if (parser.hasOption("pooling-scope")) {
            POOLING_SCOPE = getPoolingScope(parser.getOptionValue("pooling-scope"));
        }

        if (parser.hasOption("workload-type")) {
            WORKLOAD_TYPE = getWorkloadType(parser.getOptionValue("workload-type"));
        }

        if (parsedArgs.size() != 6) {
            usage("MaverickTests");
            exit(9);
        }

        HOST = parsedArgs.get(0);
        PORT = new Integer(parsedArgs.get(1)).intValue();
        USER = parsedArgs.get(2);
        PASS = parsedArgs.get(3);
        int numThreads = new Integer(parsedArgs.get(4)).intValue();
        int numFiles = new Integer(parsedArgs.get(5)).intValue();

        IConnectionPool pool = ConnectionPoolFactory.makePool();

        System.out.println("Starting MaverickTests...");
        runAll(pool, numThreads, numFiles);
        System.out.println("Done.");
        pool.flush();
    }

    private static PoolingScope getPoolingScope(String optValue) {

        for (PoolingScope mode : PoolingScope.values()) {
            if (optValue.equalsIgnoreCase(mode.toString())) {
                return mode;
            }
        }

        throw new IllegalArgumentException("Illegal value for option '--pooling-scope'");
    }

    private static WorkloadType getWorkloadType(String optValue) {

        for (WorkloadType type : WorkloadType.values()) {
            if (optValue.equalsIgnoreCase(type.toString())) {
                return type;
            }
        }

        throw new IllegalArgumentException("Illegal value for option '--workload-type'");
    }

    static class CommandParser {

        // the original args
        private final String[] args;
        // the options to be parsed
        private final List<String> options = new ArrayList<String>();
        // the option values that have been found
        private Map<String, String> optValuesFound = new HashMap<String,String>();
        // the options that have been found
        private List<String> optsFound = new ArrayList<String>();
        // the remaining args without options
        private List<String> strippedArgs = new ArrayList<String>();
        // tells if the original args already have been parsed
        private boolean parsed = false;

        public CommandParser(String[] args, String ... opts) {

            this.args = args;
            for (String opt : opts) {
                options.add(opt);
            }
        }

        private void parse() {

            if (parsed) {
                return;
            }

            parsed = true;

            for (int i = 0; i < args.length; i++) {

                boolean isOption = false;
                for (String opt : options) {
                    boolean hasValue = false;
                    String strippedOpt = opt;
                    // check for option with value
                    if (opt.endsWith(":")) {
                        hasValue = true;
                        strippedOpt = opt.substring(0, opt.length() -1);
                    }
                    String optToken;
                    if (opt.length() == 1) {
                        // single character opt is introduced by -
                        optToken = "-" + strippedOpt;
                    } else {
                        // multiple character opt is introduced by --
                        optToken = "--" + strippedOpt;
                    }

                    if (args[i].equals(optToken)) {
                        // an option was found
                        isOption = true;
                        optsFound.add(strippedOpt);
                        if (hasValue) {
                            // get the value of the option
                            if ((i+1) < args.length) {
                                optValuesFound.put(strippedOpt, args[i+1]);
                                i++;
                            } else {
                                throw new IllegalArgumentException("No value found for option '" + optToken + "'");
                            }
                        }
                    }
                }

                if (!isOption) {
                    // normal argument found
                    strippedArgs.add(args[i]);
                }
            }
        }

        /**
         * Get the value of an option.
         *
         * @param optName the option
         * @return the value or null if there was none
         */
        public String getOptionValue(String optName) {

            parse();
            return this.optValuesFound.get(optName);
        }

        /**
         * Tell if an option was present.
         *
         * @param optName the option
         * @return true if the option was present, false otherwise
         */
        public boolean hasOption(String optName) {

            parse();
            return this.optsFound.contains(optName);
        }

        /**
         * Get the arguments without any options.
         *
         * @return the args without any options
         */
        public List<String> getArgs() {

            parse();
            return this.strippedArgs;
        }
    }

    private static String getDateString() {

        GregorianCalendar cal = new GregorianCalendar();
        XMLGregorianCalendar xmlCal = DATATYPE_FACTORY.newXMLGregorianCalendar(cal);
        return xmlCal.toString();
    }

    private static void debug(int level, String message) {

        String dateString = getDateString();
        if (level <= DEBUG_LEVEL) {
            System.out.println("[" + Thread.currentThread().getName() + ", " + dateString + "] " + message);
        }
    }

    private static void workLoad(IConnectionPool pool, int taskNo, int numFiles) {

        debug(0, "-> workload(" + taskNo + ", " + numFiles + ")");
        switch(WORKLOAD_TYPE) {
            case NORMAL:
                work(pool, taskNo, numFiles);
                break;
            case NONE:
                noWork(pool, taskNo, numFiles);
                break;
            case DUMMY:
                dummyWork(pool, taskNo, numFiles);
                break;
            default:
                throw new IllegalStateException("illegal workload type '" + WORKLOAD_TYPE + "'.");
        }
        debug(0, " <- workload(" + taskNo + ", " + numFiles + ")");
    }

    private static Runnable makeRunnable(final IConnectionPool pool, final int taskNo, final int numFiles) {

        return new Runnable() {
            public void run() {

                // run the defined workload
                workLoad(pool, taskNo, numFiles);
            }
        };
    }

    private static void runAll(
            IConnectionPool pool,
            int numThreads,
            int numFiles) {

        List<Thread> threads = new ArrayList<Thread>();
        // create all threads
        for (int i = 0; i < numThreads; i++) {
            Thread thread = new Thread(makeRunnable(pool, i, numFiles));
            threads.add(thread);
        }

        // start all threads
        for (int i = 0; i < numThreads; i++) {
            Thread thread = threads.get(i);
            thread.start();
        }

        // join all threads
        for (int i = 0; i < numThreads; i++) {
            Thread thread = threads.get(i);
            try {
                thread.join();
            } catch (InterruptedException e) {
                debug(0, "Error joining thread #" + i + ": " + e);
            }
        }
    }

    private static void work(IConnectionPool pool, int taskNo, int numFiles) {

        Connection conn = null;
        try {
            // make the test directory
            String directory = ROOT_DIRECTORY + "/task-" + taskNo;
            conn = pool.create();
            makeDirectory(conn.sftpClient, directory);
            pool.destroy(conn);
            conn = null;

            // make some file names
            List<String> fileNames = new ArrayList<String>();
            for (int i = 0; i < numFiles; i++) {
                String fileName = directory + "/file-" + i;
                fileNames.add(fileName);
            }

            // write all files
            for (int i = 0; i < numFiles; i++) {
                conn = pool.create();
                String fileName = fileNames.get(i);
                InputStream is = makeContent(CONTENT_SIZE);
                TimeCounter tc = new TimeCounter();
                OutputStream os = openWrite(conn.sftpClient, fileName);
                debug(3, "Writing file " + fileName);
                write(is, os);
                os.close();
                is.close();
                debug(3, "Writing file " + fileName + " done in " + tc);
                pool.destroy(conn);
                conn = null;
            }

            // read all files
            for (int i = 0; i < numFiles; i++) {
                conn = pool.create();
                String fileName = fileNames.get(i);
                TimeCounter tc = new TimeCounter();
                InputStream is = openRead(conn.sftpClient, fileName);
                debug(3, "Reading file " + fileName);
                readAndCheck(is);
                is.close();
                debug(3, "Reading file " + fileName + " done in " + tc);
                pool.destroy(conn);
                conn = null;
            }

            // delete all files
            for (int i = 0; i < numFiles; i++) {
                conn = pool.create();
                String fileName = fileNames.get(i);
                TimeCounter tc = new TimeCounter();
                debug(3, "Deleting file " + fileName);
                delete(conn.sftpClient, fileName);
                debug(3, "Deleting file " + fileName + " done in " + tc);
                pool.destroy(conn);
                conn = null;
            }

            // delete the test directory
            conn = pool.create();
            delete(conn.sftpClient, directory);
            pool.destroy(conn);
            conn = null;
        } catch (Exception e) {
            debug(0, " ERROR: " + e.getMessage());
            e.printStackTrace();
        } finally {
            // try to cleanup connection resources
            if (conn != null) {
                disconnect(conn);
            }
        }

    }

    private static void noWork(IConnectionPool pool, int taskNo, int numFiles) {

        Connection conn = null;
        try {
            // just do the connection stuff without any real work
            for (int i = 0; i < numFiles; i++) {
                conn = pool.create();
                pool.destroy(conn);
                conn = null;
            }
        } catch (Exception e) {
            debug(0, " ERROR: " + e.getMessage());
            e.printStackTrace();
        } finally {
            // try to cleanup connection resources
            if (conn != null) {
                disconnect(conn);
            }
        }

    }

    private static void dummyWork(
            IConnectionPool pool,
            int taskNo,
            int numFiles) {

        Connection conn = null;

        try {
            // just do the connection stuff without any real work
            for (int i = 0; i < numFiles; i++) {
                conn = pool.create();
                Thread.sleep(DUMMY_SLEEP_MILLIS);
                pool.destroy(conn);
                conn = null;
            }
        } catch (Exception e) {
            debug(0, " ERROR: " + e.getMessage());
            e.printStackTrace();
        } finally {
            // try to cleanup connection resources
            if (conn != null) {
                disconnect(conn);
            }
        }

    }

    private static class Connection {

        public SshClient sshClient;
        public SftpClient sftpClient;

        public Connection(SshClient sshClient, SftpClient sftpClient) {
            this.sshClient = sshClient;
            this.sftpClient = sftpClient;
        }
    }

    private static class ConnectionPoolFactory {

        public static IConnectionPool makePool() {

            switch(POOLING_SCOPE) {
                case NONE:
                    return new NoPooling();
                case PROCESS:
                    return new ProcessConnectionPool(false);
                case PROCESS_FIFO:
                    return new ProcessConnectionPool(true);
                case THREAD:
                    return new ThreadConnectionPool();
                default:
                    throw new IllegalArgumentException("bad value for pooing scope: " + POOLING_SCOPE);
            }
        }
    }

    // Interface for a connection pool
    private interface IConnectionPool {

        /** Create a connection */
        Connection create() throws Exception;
        /** Destroy a connection */
        void destroy(Connection conn);

        /** Cleanup/flush the pool */
        void flush();
    }

    /** normal process scoped connection pooling */
    private static class ProcessConnectionPool implements IConnectionPool {

        /** the connection pool */
        private final Deque<Connection> POOL = new LinkedList<Connection>();
        private final boolean fifo;

        public ProcessConnectionPool(boolean fifo) {
            this.fifo = fifo;
        }

        public Connection create() throws Exception {

            Connection conn = null;

            // try to get a connection from the pool
            synchronized(POOL) {
                if (fifo) {
                    // using FIFO strategy we take the newest entry
                    conn = POOL.peekLast();
                    if (conn != null) {
                        POOL.removeLast();
                    }
                } else {
                    // using non-FIFO strategy we take the oldest entry
                    conn = POOL.peekFirst();
                    if (conn != null) {
                        POOL.removeFirst();
                    }
                }
            }

            // check if connection from pool is usable
            if (conn != null) {
                if (!isUsable(conn.sshClient, conn.sftpClient)) {
                    debug(0, "isUsable = false");
                    // disconnect and forget unusable connection
                    disconnect(conn);
                    conn = null;
                }
            }

            // create a fresh connection if we did not find some usable in the pool
            if (conn == null) {
                conn = connect();
            }

            // return a usable connection
            return conn;
        }

        public void destroy(Connection conn) {

            synchronized(POOL) {
                POOL.addLast(conn);
            }
        }

        public void flush() {

            synchronized(POOL) {
                Iterator<Connection> connectionIt = POOL.iterator();
                while (connectionIt.hasNext()) {
                    Connection conn = connectionIt.next();
                    connectionIt.remove();
                    disconnect(conn);
                }
            }
        }
    }

    /** thread scoped connection pooling */
    private static class ThreadConnectionPool implements IConnectionPool {

        /** the connection pool */
        private final Map<Thread, Deque<Connection>> POOL = new HashMap<Thread, Deque<Connection>>();

        public Connection create() throws Exception {

            Connection conn = null;
            Thread currentThread = Thread.currentThread();

            // try to get a connection from the pool
            synchronized(POOL) {
                Deque<Connection> connections = POOL.get(currentThread);
                if (connections != null) {
                    conn = connections.peek();
                    if (conn != null) {
                        connections.pop();
                    }
                }
            }

            // check if connection from pool is usable
            if (conn != null) {
                if (!isUsable(conn.sshClient, conn.sftpClient)) {
                    // disconnect and forget unusable connection
                    disconnect(conn);
                    conn = null;
                }
            }

            // create a fresh connection if we did not find some usable in the pool
            if (conn == null) {
                conn = connect();
            }

            // return a usable connection
            return conn;
        }

        public void destroy(Connection conn) {

            Thread currentThread = Thread.currentThread();

            synchronized(POOL) {
                Deque<Connection> connections = POOL.get(currentThread);
                if (connections == null) {
                    connections = new LinkedList<Connection>();
                    POOL.put(currentThread, connections);
                }
                connections.addLast(conn);
            }
        }

        public void flush() {

            synchronized(POOL) {
                Iterator<Thread> threadIt = POOL.keySet().iterator();
                while(threadIt.hasNext()) {
                    Thread t = threadIt.next();
                    Deque<Connection> connections = POOL.get(t);
                    Iterator<Connection> connectionIt = connections.iterator();
                    while (connectionIt.hasNext()) {
                        Connection conn = connectionIt.next();
                        connectionIt.remove();
                        disconnect(conn);
                    }
                    threadIt.remove();
                }
            }
        }

    }

    /** no connection pooling */
    private static class NoPooling implements IConnectionPool {

        public Connection create() throws Exception {

            Connection conn = connect();
            return conn;
        }

        public void destroy(Connection conn) {

            disconnect(conn);
        }

        public void flush() {
            // nothing to do
        }
    }

    /** Make a real connection */
    private static Connection connect() throws Exception {

        SshClient sshClient = null;
        SftpClient sftpClient = null;
        boolean success = false;

        try {

            TimeCounter tc = new TimeCounter();

            // get connector
            SshConnector con = SshConnector.createInstance();
            debug(2, "SshConnector.getInstance(): " + tc);

            foo(con);

            // set preferred public key
            setPreferredPublicKey(con);

            // set timeout
            setTimeout(con);

            // connect to server
            SocketTransport transport = new SocketTransport(HOST, PORT);

            //foo(transport);

            sshClient = con.connect(transport, USER);
            debug(2, "con.connect(transport, USER): " + tc);

            //foo(sshClient);

            // make proper authentication
            SshAuthentication auth = makePasswordAuthentication(PASS);

            // do the authentication
            int authResult = sshClient.authenticate(auth);
            debug(2, "sshClient.authenticate(auth): " + tc);
            if (authResult != SshAuthentication.COMPLETE) {
                throw new Exception("auth failed");
            }

            // make SFTP client
            sftpClient = makeSftpClient(sshClient);
            debug(2, "makeSftpClient(sshClient): " + tc);

            success = true;
            debug(2, "connect: " + tc);
            return new Connection(sshClient, sftpClient);
        } finally {
            if (!success) {
                cleanup(sshClient, sftpClient);
            }
        }
    }

    /** Disconnect a real connection */
    private static void disconnect(Connection conn) {

        TimeCounter tc = new TimeCounter();
        if (conn != null) {
            cleanup(conn.sshClient, conn.sftpClient);
        }
        debug(2, " disconnect: " + tc);
    }

    private static void cleanup(SshClient sshClient, SftpClient sftpClient) {

        if (sftpClient != null) {
            try {
                sftpClient.exit();
            } catch (SshException e) {
                debug(1, " ERROR: " + e.getMessage());
                e.printStackTrace();
            }
        }
        if (sshClient != null) {
            sshClient.disconnect();
        }
    }

    /**
     * Do some voodoo with the SshConnector.
     * @param con the SshConnector
     * @throws com.maverick.ssh.SshException in case of error
     */
    private static void foo(SshConnector con)
            throws SshException {

        Ssh2Context context = (Ssh2Context) con.getContext(SshConnector.SSH2);

        if (CIPHER != null) {
            context.setPreferredCipherCS(CIPHER);
            context.setPreferredCipherSC(CIPHER);
        }
        if (MAC != null) {
            context.setPreferredMacCS(MAC);
            context.setPreferredMacSC(MAC);
        }
    }

    /**
     * Do some voodoo with the SocketTransport.
     * @param transport the SocketTransport
     * @throws java.net.SocketException in case of error
     */
    @SuppressWarnings("unused")
	private static void foo(SocketTransport transport) throws SocketException {

        if (TCP_BUFFERSIZE > 0) {
            transport.setSendBufferSize(TCP_BUFFERSIZE);
            transport.setReceiveBufferSize(TCP_BUFFERSIZE);
        }

        transport.setTcpNoDelay(TCP_NODELAY);

    }

    /**
     * Set the configured preferred public key type.
     *
     * @param con the SshConnector
     * @throws SshException in case of error
     */
    private static void setPreferredPublicKey(SshConnector con)
            throws SshException {

        Ssh2Context ssh2Context = (Ssh2Context) con.getContext(SshConnector.SSH2);
        ssh2Context.setPreferredPublicKey(PREFERRED_PUBLIC_KEY_TYPE);
    }

    /**
     * Set the configured socket timeout.
     *
     * @param con the SshConnector
     * @throws SshException in case of error
     */
    private static void setTimeout(SshConnector con) throws SshException  {

        Ssh2Context ssh2Context = (Ssh2Context) con.getContext(SshConnector.SSH2);
        // UB experimental 2014-07-04
        //ssh2Context.setPartialMessageTimeout(TIMEOUT);
        ssh2Context.setSocketTimeout(TIMEOUT);
    }

    /**
     * Make the proper authentication object.
     *
     * @param password the password
     * @return the SshAuthentication object
     * @throws Exception in case of error
     */
    private static SshAuthentication makePasswordAuthentication(String password)
            throws Exception {

        PasswordAuthentication pwdAuth = new PasswordAuthentication();
        pwdAuth.setPassword(password);
        return pwdAuth;
    }

    /**
     * Make an SftpClient out of a SshClient.
     * @param client the SshClient
     * @return the SftpClient
     */
    private static SftpClient makeSftpClient(SshClient client)
            throws ChannelOpenException, SshException, SftpStatusException {

        SftpClient sftpClient;

        // suggested by Lee Painter from SSHTOOLS.
        if (client instanceof Ssh2Client) {
            Ssh2Session session = ((Ssh2Client) client).openSessionChannel(WINDOWSIZE, PACKETSIZE, null);
            session.startSubsystem("sftp");
            sftpClient = new SftpClient(session);
        } else {
            sftpClient = new SftpClient(client);
        }

        return sftpClient;
    }

    /**
     * Open the file for writing.
     *
     * @return an OutputStream for the opened file
     * @throws java.io.IOException in case of error
     */
    private static OutputStream openWrite(SftpClient sftpClient, String fileName) throws Exception {

        // use OPEN_EXCLUSIVE flag to prevent accidental file overwrite
        OutputStream os = new SftpFileOutputStream(
                sftpClient.getSubsystemChannel().openFile(
                        fileName, OPEN_CREATE | OPEN_EXCLUSIVE | OPEN_WRITE));

        return os;
    }

    private static void delete(SftpClient sftpClient, String fileName) throws Exception {

        sftpClient.rm(fileName);
    }

    private static void makeDirectory(SftpClient sftpClient, String directory)
            throws Exception {

        sftpClient.mkdirs(directory);
    }

    /**
     * Open the file for reading.
     *
     * @return an InputStream for the opened file
     * @throws java.io.IOException in case of error
     */
    private static InputStream openRead(SftpClient sftpClient, String fileName) throws Exception {

        InputStream is = sftpClient.getInputStream(fileName);
        return is;
    }

    @SuppressWarnings("unused")
	private static String makeFileName(String directory, int taskNo, int fileNo) {

        return directory + "/" + taskNo + "-" + fileNo;
    }

    /**
     * Make a test content of certain size.
     *
     * @param size the desired size
     * @return the test content
     */
    private static InputStream makeContent(int size) {

        byte bytes[] = new byte[size];
        for (int i = 0; i < size; i++) {
            bytes[i] = makeByte(i);
        }

        ByteArrayInputStream is = new ByteArrayInputStream(bytes);
        return is;
    }

    private static void write(InputStream is, OutputStream os)
    throws Exception {

        byte buffer[] = new byte[BUFFER_SIZE];
        int bytesRead;

        while ((bytesRead = is.read(buffer, 0, BUFFER_SIZE)) != -1) {
            os.write(buffer, 0, bytesRead);
        }
    }

    /**
     * Read the content from the InputStream and check if it is correct.
     *
     * @param is the InputStream
     * @throws Exception if content does not match
     */
    private static void readAndCheck(InputStream is)
    throws Exception {

        byte buffer[] = new byte[BUFFER_SIZE];
        int offset = 0;
        int bytesRead;

        while ((bytesRead = is.read(buffer, 0, BUFFER_SIZE)) != -1) {
            for (int i = 0; i < bytesRead; i++) {
                byte byteExpected = makeByte(offset);
                byte byteFound = buffer[i];
                if (byteFound != byteExpected) {
                    throw new Exception("content differs at offset " + offset);
                }
                offset++;
            }
        }

    }

    /**
     * Compute the byte value of the test content at the desired offset.
     *
     * @param offset the offset
     * @return the byte value
     */
    private static byte makeByte(int offset) {

        byte b = (byte) (offset % 256);
        return b;
    }

    // Check if connection is still usable
    public static boolean isUsable(SshClient sshClient, SftpClient sftpClient) {

        try {
            if (sshClient.isConnected()) {
                // further health check
                // will try to stat the current working directory
                String cwd = sftpClient.pwd();
                sftpClient.stat(cwd);
                return true;
            } else {
                return false;
            }
        } catch (Exception e) {
            return false;
        }
    }


    /**
     * Set the license for J2SSH Maverick.
     */
    private static void setLicense() {


    }

    /**
     * Utility class for time measurement.
     * @author ubeinges
     *
     */
    private static class TimeCounter {

        /** start time. */
        private long startTime;

        /**
         * Constructor.
         */
        public TimeCounter() {

            this.startTime = System.currentTimeMillis();
        }

        /**
         * return elapsed milliseconds.
         * @return the elapsed time in milliseconds
         */
        public long getElapsed() {

            long now = System.currentTimeMillis();
            return now - startTime;
        }

        /**
         * Reset start time to current time.
         */
        @SuppressWarnings("unused")
		public void reset() {

            this.startTime = System.currentTimeMillis();
        }

        @Override
        public String toString() {

            return this.getElapsed() + "ms";
        }
    }

}